# This should keep the .dev0 suffix in general, until a release is
# made.  The pre-release makefile target takes care of stripping the
# .dev0 out.  The post-release target increments the version and adds
# .dev0 back.
__version__ = '101.0.11'
