################################################################################
# Copyright (c) 2006-2017 Franz Inc.
# All rights reserved. This program and the accompanying materials are
# made available under the terms of the MIT License which accompanies
# this distribution, and is available at http://opensource.org/licenses/MIT
################################################################################

# Just an alias, to avoid some circular imports without breaking
# backwards compatibility.
from franz.openrdf.query.repositoryresult import *
